import numpy as np
import pandas as pd

from enlopy.utils import make_timeseries , clean_convert, _freq_to_sec


class Test_make_timeseries():

    def test_ndarray_h(self):
        a = np.random.rand(8760)
        b = make_timeseries(a)
        assert isinstance(b, pd.Series)
        assert np.isclose(sum(a), sum(b))

    def test_ndarray_15m(self):
        a = np.random.rand(35040)
        b = make_timeseries(a)
        assert isinstance(b, pd.Series)
        assert np.isclose(sum(a), sum(b))

    def test_ndarray_x(self):
        a = np.random.rand(8730)
        b = make_timeseries(a, freq='h')
        assert isinstance(b, pd.Series)
        assert np.isclose(sum(a), sum(b))

    def test_pdseries(self):
        a = pd.Series(np.random.rand(8760))
        b = make_timeseries(a)
        assert isinstance(b, pd.Series)
        assert np.isclose(sum(a), sum(b))

    def test_pddataframe(self):
        a = pd.DataFrame(np.random.rand(8760))
        b = make_timeseries(a)
        assert isinstance(b, pd.DataFrame)

    def test_pddataframe_2d(self):
        a = pd.DataFrame(np.random.rand(35040,3))
        b = make_timeseries(a)
        assert isinstance(b, pd.DataFrame)

    def test_2d_ndarray(self):
        a = np.random.rand(8760,5)
        b = make_timeseries(a)
        assert isinstance(b, pd.DataFrame)

    def test_empty_frame(self):
        a = np.array([])
        b = make_timeseries(a, freq='h')
        assert isinstance(b, pd.Series) and len(b)==0

    def test_empty_frame_to_indexed_empty(self):
        b = make_timeseries(freq='h', length=8760)
        assert isinstance(b, pd.Series) and len(b) == 8760

    def test_multiannual_timeseries(self):
        a = np.random.rand(8760*2)
        b = make_timeseries(a, freq='h')
        assert isinstance(b, pd.Series)
        assert np.isclose(sum(a), sum(b))


class Test_clean_convert():

    def test_ndarray_to_series_indexed(self):
        a = np.random.rand(8760)
        b = clean_convert(a, force_timed_index=True, always_df=False)
        assert isinstance(b, pd.Series) and isinstance(b.index, pd.DatetimeIndex)

    def test_ndarray_to_df_indexed(self):
        a = np.random.rand(8760)
        b = clean_convert(a, force_timed_index=True, always_df=True)
        assert isinstance(b, pd.DataFrame) and isinstance(b.index, pd.DatetimeIndex)

    def test_1d_series_to_frame(self):
        a = pd.Series(np.random.rand(8760))
        b = clean_convert(a, force_timed_index=True, always_df=True)
        assert isinstance(b, pd.DataFrame) and isinstance(b.index, pd.DatetimeIndex)

    def test_2d_ndarray_to_df_indexed(self):
        a = np.random.rand(8760, 2)
        b = clean_convert(a, force_timed_index=True, always_df=True)
        assert isinstance(b, pd.DataFrame) and isinstance(b.index, pd.DatetimeIndex)

    def test_list_to_series(self):
        a = list(np.random.rand(8760))
        b = clean_convert(a, force_timed_index=True, always_df=False)
        assert isinstance(b, pd.Series)


def test_freq_to_sec():
    a = _freq_to_sec('d')
    assert a == 60 * 60 * 24


# -*- coding: utf-8 -*-
"""
Methods that generate or adjusted energy related timeseries based on given assumptions/input
"""

import numpy as np
import pandas as pd
import scipy.interpolate
import scipy.linalg
import scipy.stats

from .utils import make_timeseries, clean_convert
from .analysis import countweekend_days_per_month

__all__ = ['disag_upsample', 'gen_daily_stoch_el', 'gen_load_from_daily_monthly', 'gen_load_sinus', 'gen_load_from_LDC',
           'gen_load_from_PSD', 'gen_gauss_markov', 'remove_outliers', 'gen_demand_response', 'add_noise',
           'gen_corr_arrays', 'gen_analytical_LDC']

_EPS = np.finfo(np.float64).eps

def disag_upsample(Load, disag_profile, to_offset='h'):
    """ Upsample given timeseries, disaggregating based on given load profiles.
    e.g. From daily to hourly. The load of each day is distributed according to the disaggregation profile. The sum of each day remains the same.

    Arguments:
        Load (pd.Series): Load profile to disaggregate
        disag_profile (pd.Series, np.ndarray): disaggregation profile to be used on each timestep of the load. Has to be compatible with selected offset.
        to_offset (str): Resolution of upsampling. has to be a valid pandas offset alias. (check `here <http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases>`__ for all available offsets)
    Returns:
        pd.Series: the upsampled timeseries
   """
    #First reindexing to the new resolution.
    orig_freq = Load.index.freqstr
    start = Load.index[0]
    end = Load.index[-1] + 1 * Load.index.freq #An extra period is needed at the end to match the sum FIXME
    df1 = Load.reindex(pd.date_range(start, end, freq=to_offset, inclusive='left'))

    def mult_profile(x, profile):
        #Normalizing to keep the sum the same..
        profile = profile / np.sum(profile)
        return x.mean() * profile  #using mean() assuming that there is one value and the rest is nan

    #then transform per sampled period correspnding to the len(disag_profile)
    return df1.resample(orig_freq).transform(mult_profile, disag_profile).dropna()


def gen_daily_stoch_el(total_energy=1.0):
    """Generate stochastic dummy daily load based on hardcoded values. These values are the result of
    statistical analysis of electric loads profiles of more than 100 households. Mean and standard deviations per timestep were extracted
    from the normalized series. These are fed to :meth:`gen_gauss_markov` method.
    Arguments:
        total_energy: Sum of produced timeseries (daily load)
    Returns:
        nd.array: random realization of timeseries
    """

    means = np.array([0.02603978, 0.02266633, 0.02121337, 0.02060187, 0.02198724,
             0.02731497, 0.03540281, 0.0379463, 0.03646055, 0.03667756,
             0.03822946, 0.03983243, 0.04150124, 0.0435474, 0.0463219,
             0.05051979, 0.05745442, 0.06379564, 0.06646279, 0.06721004,
             0.06510399, 0.05581182, 0.04449689, 0.03340142])
    stds = np.array([0.00311355, 0.00320474, 0.00338432, 0.00345542, 0.00380437,
           0.00477251, 0.00512785, 0.00527501, 0.00417598, 0.00375874,
           0.00378784, 0.00452212, 0.00558736, 0.0067245, 0.00779101,
           0.00803175, 0.00749863, 0.00365208, 0.00406937, 0.00482636,
           0.00526445, 0.00480919, 0.00397309, 0.00387489])
    a = gen_gauss_markov(means,stds, .8)
    return a / a.sum() * total_energy


def gen_load_from_daily_monthly(ML, DWL, DNWL, weight=0.5, year=2015):
    """Generate annual timeseries using monthly demand and daily profiles.
    Working days and weekends are built from different profiles having different weighting factors.

    Arguments:
        ML: monthly load (size = 12)
        DWL: daily load (working day) (size = 24). Have to be normalized (sum=1)
        DNWL: daily load (non working day) (size = 24) Have to be normalized (sum=1)
        weight: weighting factor between working and non working day (0 - 1)
    Returns:
        pd.Series: Generated timeseries
    """
    #TODO: refactor. Can i use disag_upsample() ?
    if not(np.isclose(DWL.sum(), 1) and np.isclose(DNWL.sum(), 1)):
        raise ValueError('Daily profiles should be normalized')
        #TODO: Normalize here?
    out = make_timeseries(year=year, length=8760, freq='h')  # Create empty pandas with datetime index
    import calendar
    febdays = 29 if calendar.isleap(year) else 28
    Days = np.array([31, febdays, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])
    # Assumptions for non working days per month. Only weekends
    # TODO: Custom Calendars with holidays BDays
    DaysNW = countweekend_days_per_month(out.resample('D').mean())
    DaysW = Days - DaysNW
    for month in range(12):
        # Estimate total load for working and non working day
        TempW = (ML[month] * weight * DaysW[month] /
                (weight * DaysW[month] + (1 - weight) * DaysNW[month]) / DaysW[month] )
        TempNW = (ML[month] * (1 - weight) * DaysNW[month] /
                 (weight * DaysW[month] + (1 - weight) * DaysNW[month]) / DaysNW[month])
        for hour in range(24):
            out.loc[(out.index.month == month + 1) &  #months dont start from 0
                   (out.index.weekday < 5) &
                   (out.index.hour == hour)] = (TempW * DWL[hour])
            out.loc[(out.index.month == month + 1) &
                   (out.index.weekday >= 5) &
                   (out.index.hour == hour)] = (TempNW * DNWL[hour])
    return out


def gen_load_sinus(daily_1, daily_2, monthly_1, monthly_2, annually_1, annually_2):
    """ Generate sinusoidal load with daily, weekly and yearly seasonality. Each term is estimated based
    on the following expression: :math:`f(x;A1,A2,w) =  A1 \\cos(2 \\pi/w \\cdot x) + A2 \\sin(2 \\pi/w \\cdot x)`

    Arguments:
        daily_1 (float): cosine coefficient for daily component (period 24)
        daily_2 (float): sinus coefficient for daily component (period 24)
        monthly_1 (float): cosine coefficient for monthly component (period 168)
        monthly_2 (float): sinus coefficient for monthly component (period 168)
        annually_1 (float): cosine coefficient for annual component (period 8760)
        annually_2 (float): sinus coefficient for annual component (period 8760)
    Returns:
        pd.Series: Generated timeseries
    """
    def sinusFunc(x, w, A1, A2): # fourrier coefficient
        return A1 * np.cos(2 * np.pi/w * x) + A2 * np.sin(2 * np.pi/w * x)

    x = np.arange(0, 8760)
    # daily, weekly, annual periodicity     #TODO: custom periodicity
    coeffs ={24: (daily_1, daily_2),
             168: (monthly_1, monthly_2),
             8760: (annually_1, annually_2)
             }

    out = 0
    for period, values in coeffs.items():
        out += sinusFunc(x, period, *values)

    return make_timeseries(out)


def gen_corr_arrays(Na, length, M, to_uniform=True):
    """ Generating correlated normal variates.
    Assume one wants to create a vector of random variates Z which is
    distributed according to Z~N(μ,Σ) where μ is the vector of means,
    and Σ is the variance-covariance matrix.
    http://comisef.wikidot.com/tutorial:correlateduniformvariates

    Arguments:
        Na (int): number of vectors e.g (3)
        length (int): generated vector size (e.g 8760)
        M (np.ndarray): correlation matrix. Should be of size Na x Na
        to_uniform (bool): True if the correlation matrix needs to be adjusted for uniforms
    Returns:
         np.ndarray: Realization of randomly generated correlated variables. Size : (Na, length) e.g. (3, 8760)
    """
    if Na != np.size(M, 0):  # rows of pars have to be the same size as rows and cols of M
        print('Parameters and corr matrix dimensions do not agree.')
        return False

    newM = M.copy()  # changing an array element without copying it changes it globally!
    u = np.random.randn(length, Na)
    if min(np.linalg.eig(M)[0]) < 0:  # is M positive definite?
        print ('Error: Eigenvector is not positive. Trying to make positive, but it may differ from the initial.')
        # Make M positive definite:
        la, v = np.linalg.eig(newM)
        la[la < 0] = np.spacing(1)  # Make all negative eigenvalues zero
        ladiag = np.diag(la)  # Diagonal of eigenvalues
        newM = np.dot(np.dot(v, ladiag), v.T)  # Estimate new M = v * L * v'

    # Transformation is needed to change normal to uniform (Spearman - Pearson)
    if to_uniform:
        for i in np.arange(0, Na):  # 1:Na
            for j in np.arange(max(Na-1, i), Na):  # max(Na-1,i):Na
                if i != j:
                    newM[i, j] = 2 * np.sin(np.pi * newM[i, j] / 6)
                    newM[j, i] = 2 * np.sin(np.pi * newM[j, i] / 6)

    if min(np.linalg.eig(newM)[0]) <= 0:
        print ('Error: Eigenvector is still not positive. Aborting')
        return False

    cF = scipy.linalg.cholesky(newM)
    Y = np.dot(u, cF).T
    Y = scipy.stats.norm.cdf(Y)  # remove if you produce random.rand?
    return Y


def gen_load_from_LDC(LDC, Y=None, N=8760):
    """ Generate loads based on a Inverse CDF, such as a Load Duration Curve (LDC)
    Inverse transform sampling: Compute the value x such that F(x) = u.
    Take x to be the random number drawn from the distribution described by F.

    .. note::
        Due to the sampling process this function produces load profiles with unrealistic temporal sequence, which means that they cannot be treated as
        timeseries. It is recommended that :meth:`gen_load_from_PSD` is used afterwards.

    Arguments:
        LDC (np.ndarray): Load duration curve (2 x N) vector of the x, y coordinates of an LDC function (results of (get_LDC).
            x coordinates have to be normalized (max: 1 => 8760hrs )
        Y (nd.array): a vector of random numbers. To be used for correlated loads.
            If None is supplied a random vector (8760) will be created.
        N (int): Length of produced timeseries (if Y is not provided)

    Returns:
         np.ndarray: vector with the same size as Y that respects the statistical distribution of the LDC
    """
    if Y is None:  # if there is no Y, generate a random vector
        Y = np.random.rand(N)
    func_inv = scipy.interpolate.interp1d(LDC[0], LDC[1], bounds_error=False, fill_value=0)
    simulated_loads = func_inv(Y)
    # ------- Faster way:   # np.interp is faster but have to sort LDC
    # if np.all(np.diff(LDC[0]) > 0) == False: #if sorted
    #idx = np.argsort(LDC[0])
    #LDC_sorted = LDC[:, idx].copy()
    #simulated_loads = np.interp(Y, LDC_sorted[0], LDC_sorted[1])
    # no need to insert timed index since there is no spectral information
    return simulated_loads


def gen_load_from_PSD(Sxx, x, dt=1):
    """
    Algorithm for generating samples of a random process conforming to spectral
    density Sxx(w) and probability density function p(x).

    .. note::
        This is done by an iterative process which 'shuffles' the timeseries till convergence of both
        power spectrum and marginal distribution is reached.
        Also known as "Iterated Amplitude Adjusted Fourier Transform (IAAFT). Adopted from `J.M. Nichols, C.C. Olson, J.V. Michalowicz, F. Bucholtz, (2010), "A simple algorithm for generating spectrally colored, non-Gaussian signals" Probabilistic Engineering Mechanics, Vol 25, 315-322`
        and `Schreiber, T. and Schmitz, A. (1996) "Improved Surrogate Data for Nonlinearity Tests", Physical Review Letters, Vol 77, 635-638.`

    Arguments:
        Sxx: Spectral density (two sided)
        x: Sequence of observations created by the desirable PDF. You can use :meth:`gen_load_from_LDC` for that.
        dt: Desired temporal sampling interval. [Dt = 2pi / (N * Dw)]
    Returns:
        pd.Series: The spectrally corrected timeseries

    """
    N = len(x)
    Sxx[int(N/2)+1] = 0  # zero out the DC component (remove mean)
    Xf = np.sqrt(2 * np.pi * N * Sxx / dt)  # Convert PSD to Fourier amplitudes
    Xf = np.fft.ifftshift(Xf)  # Put in Matlab FT format
    # The following lines were commented out because they outscale the data
    # modifying thus its PDF. However, according to Nichols et al. they
    # guarantee that the new data match the signal variance
    #vs = (2 * np.pi / N / dt) * sum(Sxx) * (N / (N-1))  # Get signal variance (as determined by PSD)
    #out = x * np.sqrt(vs / np.var(x))
    out = x
    mx = np.mean(out)
    out = out - mx  # subtract the mean
    indx = np.argsort(out)
    xo = out[indx].copy()  # store sorted signal xo with correct p(x)

    k = 1
    indxp = np.zeros(N)  # initialize counter
    while(k):
        Rk = np.fft.fft(x)  # Compute FT
        Rp = np.angle(Rk)  # ==> np.arctan2(np.imag(Rk), np.real(Rk))  # Get phases
        out = np.real(np.fft.ifft(np.exp(1j * Rp) * np.abs(Xf)))  # Give signal correct PSD
        indx = np.argsort(out)  # Get rank of signal with correct PSD
        out[indx] = xo  # rank reorder (simulate nonlinear transform)
        k = k + 1  # increment counter
        if np.array_equal(indx, indxp):
            print('Converged after {} iterations'.format(k))
            k = 0  # if we converged, stop
        indxp = indx  # re-set ordering for next iter
    out = out + mx  # Put back in the mean
    return out


def gen_gauss_markov(mu, st, r):
    """ Generate timeseries based on means, stadnard deviation and autocorrelation per timestep

    .. note::
        Based on `A.M. Breipohl, F.N. Lee, D. Zhai, R. Adapa, A Gauss-Markov load model for the application in risk evaluation
        and production simulation, Transactions on Power Systems, 7 (4) (1992), pp. 1493-1499`

    Arguments:
        mu: array of means. Can be either 1d or 2d
        st: array of standard deviations. Can be either 1d or 2d. Can be either scalar (same for entire timeseries or array with the same length as the timeseries
        r:  Autoregressive coefficient AR(1). Has to be between  [-1,1]. Can be either scalar (same for entire timeseries or array with the same length as the timeseries
    Returns:
        pd.Series, pd.DataFrame: a realization of the timeseries
    """
    mu = np.atleast_2d(mu)
    loadlength = mu.shape
    rndN = np.random.randn(*loadlength)
    if np.atleast_2d(st).shape[1] == 1:
        noisevector = st * np.ones(loadlength)
    elif len(st) == loadlength[1]:
        noisevector =  np.atleast_2d(st)
    else:
        raise ValueError('Length of standard deviations must be the same as the length of means. You can also use one value for the entire series')

    if np.atleast_2d(r).shape[1] == 1:
        rvector = r * np.ones(loadlength)
    elif len(r) == loadlength[1]:
        rvector = np.atleast_2d(r)
    else:
        raise ValueError('Length of autocorrelations must be the same as the length of means. You can also use one value for the entire series')

    y = np.zeros(loadlength)
    noisevector[noisevector == 0] = _EPS
    y[:,0] = mu[:,0] + noisevector[:, 0] * rndN[:, 0]

 #   for t in mu.T:
    for i in range(mu.shape[1]):
        y[:,i] = (mu[:,i] +
                r * noisevector[:, i] /
                noisevector[:, i - 1] * (y[:, i - 1] - mu[:, i - 1]) +
                noisevector[:, i] * np.sqrt(1 - rvector[:, i] ** 2) * rndN[:, i])
    return y.squeeze()


def add_noise(Load, mode, st, r=0.9, Lmin=0):
    """ Add noise with given characteristics.

    Arguments:
        Load (pd.Series,pd.DataFrame): 1d or 2d timeseries
        mode (int):1 Normal Distribution, 2: Uniform Distribution, 3: Gauss Markov (autoregressive gaussian)
        st (float): Noise parameter. Scaling of random values
        r (float): Applies only for mode 3. Autoregressive coefficient AR(1). Has to be between  [-1,1]
        Lmin (float): minimum load values. This is used to trunc values below zero if they are generated with a lot of noise
    Returns:
        pd.Series: Load with noise
    """

    L = np.atleast_2d(Load)

    if st == 0:
        print('No noise to add')
        return Load

    loadlength = L.shape  # 8760
    if mode == 1:  # Normal
        noisevector = st * np.random.randn(*loadlength)  # gauss.. should it have a zero sum?
        out = L * (1 + noisevector)
    elif mode == 2:  # Uniform
        noisevector = st * np.random.rand(*loadlength)
        out = L * ((1 - st) + st * noisevector)
    elif mode == 3:  # Gauss-Markov, same as
        out = gen_gauss_markov(L, st, r)
    else:
        raise ValueError('Not available mode')
    out[out < Lmin] = Lmin  # remove negative elements
    return clean_convert(np.squeeze(out), force_timed_index=False) # assume hourly timeseries if no timeindex is passed


def gen_analytical_LDC(U, duration=8760,bins=1000):
    r"""Generates the Load Duration Curve based on empirical parameters. The following equation is used.
    :math:`f(x;P,CF,BF) = \\frac{P-x}{P-BF \\cdot P}^{\\frac{CF-1}{BF-CF}}`

    Arguments:
        U (tuple): parameter vector [Peak load, capacity factor%, base load%, hours] or dict
    Returns:
        np.ndarray: a 2D array [x, y] ready for plotting (e.g. plt(\*gen_analytical_LDC(U)))
    """
    if isinstance(U, dict):
        P = U['peak']  # peak load
        CF = U['LF']  # load factor
        BF = U['base']  # base load
        h = U['hourson']  # hours
    else: #unpack
        P, CF, BF, h = U

    x = np.linspace(0, P, bins)

    ff = h * ((P - x)/(P - BF * P))**((CF - 1)/(BF - CF))
    ff[x < (BF*P)] = h
    ff[x > P] = 0
    return ff/duration, x


def gen_demand_response(Load, percent_peak_hrs_month=0.03, percent_shifted=0.05, shave=False):
    """Simulate a demand response mechanism that makes the load profile less peaky.
    The load profile is analyzed per selected period (currently month) and the peak hours have their load shifted
    to low load hours or shaved. When not shaved the total load is the same as that one from the initial timeseries,
    otherwise it is smaller due to the shaved peaks. The peak load is reduced by a predefined percentage.

    Arguments:
        Load (pd.Series): Load
        percent_peak_hrs_month (float): fraction of hours to be shifted
        percent_shifted (float): fraction of energy to be shifted if the day is tagged for shifting/shaving
        shave (bool): If False peak load will be transfered to low load hours, otherwise it will be shaved.
    Return:
        pd.Series: New load profile with reduced peaks. The peak can be shifted to low load hours or shaved
    """
    if not isinstance(Load.index, pd.DatetimeIndex):
        print ('Need date Time indexed series. Trying to force one.')
        Load = clean_convert(Load, force_timed_index=True)
    demand = Load

    def hours_per_month(demand):
        """Assign to each row hours per month"""
        dic_hours_per_month = demand.groupby(demand.index.month).count().to_dict()
        return demand.resample('M').transform(lambda x: list(map(dic_hours_per_month.get, x.index.month)))

    # Monthly demand rank
    # TODO: parametrize: we can check peaks on a weekly or daily basis
    demand_m_rank = demand.resample('M').transform(lambda x: x.rank(method='min', ascending=False))

    # find which hours are going to be shifted
    bool_shift_from = demand_m_rank <= np.round(hours_per_month(demand) * percent_peak_hrs_month)

    DR_shift_from = percent_shifted * demand  # demand_fpeak * total_demand * percent_shifted
    DR_shift_from[~bool_shift_from] = 0

    # find hours that are going to have
    if shave:
        #If (peak) shaving we do not shift the loads anywhere
        DR_shift_to = 0
    else:
        # Estimate amount of load to be shifted per month
        sum_shifted = DR_shift_from.groupby(DR_shift_from.index.month).sum()
        count_shifted = DR_shift_from[DR_shift_from > 0].groupby(DR_shift_from[DR_shift_from > 0].index.month).count()
        shift_to_month = sum_shifted / count_shifted
        #Find which hours are going to be filled with the shifted load
        bool_shift_to = demand_m_rank > np.round(hours_per_month(demand) * (1 - percent_peak_hrs_month))

        df_month = pd.Series(demand.index.month, index=demand.index)
        DR_shift_to = df_month.map(shift_to_month)
        DR_shift_to[~bool_shift_to] = 0

    # Adjusted hourly demand
    dem_adj = demand.copy()
    dem_adj[bool_shift_from] = dem_adj[bool_shift_from] * (1 - percent_shifted)
    dem_adj[~bool_shift_from] = dem_adj[~bool_shift_from] + DR_shift_to

    # In case of load shift check that the sum of initial timeseries is similar to the reshaped one
    if not np.isclose(dem_adj.sum(), Load.sum()):
        raise ValueError('Sum is not the same. Probably you overdid it with the shifting parameters.'
                         'Please try with more conservative ones.')
    return dem_adj


def remove_outliers(Load, **kwargs):
    """ Removes outliers identified by :meth:`detect_outliers` and replaces them by interpolated value.

    Arguments:
        Load: input timeseries
        **kwargs: Exposes keyword arguments of :meth:`detect_outliers`
    Returns:
        Timeseries cleaned from outliers
    """
    from .analysis import detect_outliers
    outlier_idx = detect_outliers(Load, **kwargs)
    filtered_series = Load.copy()
    filtered_series[outlier_idx] = np.nan
    return filtered_series.interpolate(method='time')



import numpy as np
import pandas as pd

from .utils import clean_convert

__all__ = ['reshape_timeseries', 'get_LDC', 'get_load_archetypes', 'get_load_stats', 'detect_outliers']

def reshape_timeseries(Load, x='dayofyear', y=None, aggfunc='sum'):
    """Returns a reshaped pandas DataFrame that shows the aggregated load for selected
    timeslices. e.g. time of day vs day of year

    Parameters:
        Load (pd.Series, np.ndarray): timeseries
        x (str): x axis aggregator. Has to be an accessor of pd.DatetimeIndex
         (year, dayoftime, week etc.)
        y (str): similar to above for y axis
    Returns:
        reshaped pandas dataframe according to x,y
    """

    # Have to convert to dataframe in order for pivottable to work
    # 1D, Dataframe
    a = clean_convert(Load.copy(), force_timed_index=True, always_df=True)
    a.name = 0
    if len(a.columns) > 1:
        raise ValueError('Works only with 1D')

    if x is not None:
        if x == 'week':  # For a strange reason pandas moved week accesor to isocalendar
            a[x] = a.index.isocalendar()['week']
        else:
            a[x] = getattr(a.index, x)
    if y is not None:
        if y == 'week':
            a[y] = a.index.isocalendar()['week']
        else:
            a[y] = getattr(a.index, y)
    a = a.reset_index(drop=True)

    return a.pivot_table(index=x, columns=y,
                         values=a.columns[0],
                         aggfunc=aggfunc).T


def get_LDC(Load, x_norm=True, y_norm=False):
    r"""Generates the Load Duration Curve based on a given load. For 2-dimensional dataframes the x-axis sorting is done based on sum of all series. Sorting on the y-axis is done based on the coefficient of variance.

    Arguments:
        Load (pd.Series): timeseries
        x_norm (bool): Normalize x axis (0,1)
        y_norm (bool): Normalize y axis (0,1)
    Returns:
        np.ndarray: tuple (x, y) ready for plotting (e.g. plt(\*LDC_load(load)))
    """
    Load1 = clean_convert(Load,force_timed_index=False)
    if Load1.ndim >= 2:
        # Sort x axis by total value
        sorted_ind = Load1.sum(axis=1).sort_values(ascending=False).index
        # Sort series by variance coefficient. Baseline load should have smaller CV and should go lower
        sorted_cols = (Load1.std()/Load1.mean()).sort_values().index
        y = Load1.loc[sorted_ind, sorted_cols].values
    else:
        y = Load1.sort_values(ascending=False).values
    x = np.arange(1, len(y) + 1 )
    if x_norm:
        x = x / len(x)
    if y_norm:
        y = y / y.max()
    return x, y

    # remove nan because histogram does not work
    #load_masked = Load[~np.isnan(Load)]
    #n, xbins = np.histogram(load_masked, bins=bins, density=True)
    # xbins = xbins[:-1] #remove last element to make equal size
    #cum_values = np.zeros(xbins.shape)
    #cum_values[1:] = np.cumsum(n*np.diff(xbins))
    #out = np.array([1-cum_values, xbins])
    # out = np._r[[1 0], out] # Add extra point
    #if trunc_0: # Trunc non zero elements
    #    out[out < 0] = 0
    #return out


def get_load_archetypes(Load, k=2, x='hour', y='dayofyear', plot_diagnostics=False):
    """Extract typical load profiles using k-means and vector quantization. the time scale of archetypes depend on the selected dimensions (x,y).
    For the default values daily archetypes will be extracted.

    Parameters:
        Load (pd.Series): timeseries
        k (int): number of archetypes to identify and extract
        x (str): This will define how the timeseries will be grouped by. Has to be an accessor of pd.DatetimeIndex
        y (str): similar to above for y axis.
        plot_diagnostics (bool): If true a figure is plotted showing an overview of the results
    Returns:
        np.ndarray: dimensions (k, len(x))
    """
    from scipy.cluster.vq import whiten, kmeans, vq

    df = reshape_timeseries(Load, x=x, y=y, aggfunc='mean').astype(float)
    df_white = whiten(df)
    clusters_center, __ = kmeans(df_white, k)
    clusters_center_dewhitened = clusters_center.T * np.array([df.std(), ] * k ).T

    if plot_diagnostics:
        try:
            import matplotlib.pyplot as plt
            clusters, _ = vq(df_white, clusters_center)
            cm = _n_colors_from_colormap(k)
            ax1 = df.T.plot(legend=False, alpha=.1,
                            color=[cm[i] for i in clusters])
            # Add colored cluster centers as lines
            ax1.set_prop_cycle('color', cm)
            ax1.plot(clusters_center_dewhitened, linewidth=3, linestyle='--')
            plt.figure()  # FIXME: works only with weekdays
            day_clusters = pd.DataFrame({y: Load.resample('d').mean().index.weekday,
                                         'clusters': clusters,
                                         'val': 1})
            x_labels = "Mon Tue Wed Thu Fri Sat Sun".split()
            day_clusters.pivot_table(columns=y, index='clusters',
                                     aggfunc='count').T.plot.bar(stacked=True)
            plt.gca().set_xticklabels(x_labels)
        except Exception: #FIXME: specify exception
            print ('Works only with daily profile clustering')

    return clusters_center_dewhitened


def get_load_stats(Load, per='Y') -> pd.DataFrame:
    """Find load profile characteristics. Among other it estimates: peak, load factor, base load factor, operating hours,

    Arguments:
        Load: timeseries of load to be examined. A timeseries index is needed.
        per: reporting periods. Annual by default. Based on pandas time offsets
    Returns:
         dict: Parameter dictionary
    """
     #TODO 2D
    from .stats import all_stats_desc

    Load1 = clean_convert(Load, force_timed_index=True, freq='h')

    g = Load1.groupby(pd.Grouper(freq=per))
    if len(g) > 100:
        print ('Warning: Too many periods ({}) selected'.format(len(g)))
    p_dict = {}
    for period, load_per in g:
        ind = str(period.to_period(freq=per))
        p_dict[ind] = {k: v(load_per) for k, v in all_stats_desc.items()}  #  named tuple instead of dict?
    return pd.DataFrame.from_dict(p_dict)


def _n_colors_from_colormap(n, cmap='Set1'):
    """ Returns lists of color tuples(RGBA)
        n: number of colours
        cmap: matplotlib colormap
    """
    from matplotlib.cm import get_cmap
    cm = get_cmap(cmap)
    return [cm(1.*i/n) for i in range(n)]

def detect_outliers(Load, threshold=None, window=5, plot_diagnostics=False):
    """ Detect and optionally remove outliers based on median rolling window filtering.
    Inspired by https://ocefpaf.github.io/python4oceanographers/blog/2015/03/16/outlier_detection/

    Arguments:
        Load: input timeseries
        threshold: if None then 3 sigma is selected as threshold
        window: how many values to check
        plot_diagnostics: Plot diagnostics to check whether the outliers were removed accurately
    Return:
        index position of detected outliers
   """
    # TODO : Clean zero values (interpolate)

    a = Load.rolling(window=window, center=True).median().fillna(method='bfill').fillna(method='ffill')
    difference = np.abs(Load - a)

    if threshold is None:
        threshold = 3 * np.std(Load)
    outlier_idx = difference > threshold
    if plot_diagnostics:
        if len(outlier_idx > 0):
            kw = dict(marker='o', linestyle='none', color='r', alpha=0.5)
            Load.plot()
            Load[outlier_idx].plot(**kw)
        else:
            print('No outliers detected. If you think that there are, try to raise the threshold')
    return outlier_idx


def countweekend_days_per_month(df, weekdays=True):     #TODO generalize count_x_per_y
    """Count number of occurrences where the day is Saturday or Sunday. Loops per month"""
    from collections import Counter
    out = []
    for __, values in df.index.groupby(df.index.month).items():
        m_count = Counter([day.weekday() for day in values])
        out.append(m_count[5] + m_count[6]) # 5,6 is saturday sunday
    return out


import matplotlib.pyplot as plt
import matplotlib.cm as cm
import pandas as pd

import numpy as np

from .analysis import reshape_timeseries, clean_convert, get_LDC

__all__ = ['plot_heatmap', 'plot_3d', 'plot_percentiles', 'plot_rug', 'plot_boxplot', 'plot_LDC' ]

def plot_heatmap(Load, x='dayofyear', y='hour', aggfunc='sum', bins=8,
                figsize=(16,6), edgecolors='none', cmap='Oranges', colorbar=True, ax=None, **pltargs):
    """ Returns a 2D heatmap of the reshaped timeseries based on x, y

    Arguments:
        Load: 1D pandas with timed index
        x: Parameter for :meth:`enlopy.analysis.reshape_timeseries`
        y: Parameter for :meth:`enlopy.analysis.reshape_timeseries`
        bins: Number of bins for colormap
        edgecolors: colour of edges around individual squares. 'none' or 'w' is recommended.
        cmap: colormap name (from colorbrewer, matplotlib etc.)
        **pltargs: Exposes matplotlib.plot arguments
    Returns:
        2d heatmap
    """
    x_y = reshape_timeseries(Load, x=x, y=y, aggfunc=aggfunc)
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    cmap_obj = cm.get_cmap(cmap, bins)
    heatmap = ax.pcolor(x_y, cmap=cmap_obj, edgecolors=edgecolors, **pltargs)
    if colorbar:
        fig.colorbar(heatmap)
    ax.set_xlim(right=len(x_y.columns))
    ax.set_ylim(top=len(x_y.index))
    ax.set_xlabel(x)
    ax.set_ylabel(y)


def plot_3d(Load, x='dayofyear', y='hour', aggfunc='sum', bins=15,
           cmap='Oranges', colorbar=True, **pltargs):
    """ Returns a 3D plot of the reshaped timeseries based on x, y

    Arguments:
        Load: 1D pandas with timed index
        x: Parameter for :meth:`enlopy.analysis.reshape_timeseries`
        y: Parameter for :meth:`enlopy.analysis.reshape_timeseries`
        bins: Number of bins for colormap
        cmap: colormap name (from colorbrewer, matplotlib etc.)
        **pltargs: Exposes :meth:`matplotlib.pyplot.surface` arguments
    Returns:
        3d plot
    """
    import mpl_toolkits.mplot3d  # necessary for orojection=3d

    x_y = reshape_timeseries(Load, x=x, y=y, aggfunc=aggfunc)

    fig = plt.figure(figsize=(12, 6))
    ax = fig.add_subplot(111, projection='3d')
    cmap_obj = cm.get_cmap(cmap, bins)
    X, Y = np.meshgrid(range(len(x_y.columns)), range(len(x_y.index)))
    surf = ax.plot_surface(X, Y, x_y, cmap=cmap_obj, rstride=1, cstride=1,
                           shade=False, antialiased=True, lw=0, **pltargs)
    if colorbar:
        fig.colorbar(surf)
    # Set viewpoint.
    # ax.azim = -130
    ax.elev = 45
    ax.auto_scale_xyz([0, len(x_y.columns)],
                      [0, len(x_y.index)],
                      [0, x_y.max().max()])
    ax.set_xlabel(x)
    ax.set_ylabel(y)


def plot_percentiles(Load, x='hour', zz='week', perc_list=[[5, 95], [25, 75], 50], ax=None, color='blue', **kwargs):
    """Plot predefined percentiles per timestep

    Arguments:
        Load: 1D pandas with timed index
        x (str): x axis aggregator. See :meth:`enlopy.analysis.reshape_timeseries`
        zz (str): similar to above for y axis
        perc_list(list): List of percentiles to plot. If it is an integer then it will be plotted as a line. If it is list it has to contain two items and it will be plotted using fill_between()
        **kwargs: exposes arguments of :meth:`matplotlib.pyplot.fill_between`
    Returns:
        Plot

    """
    if ax is None: # Hack for nice jupyter notebook compatibility
        ax=plt.gca()
    a = reshape_timeseries(Load, x=x, y=zz, aggfunc='mean')
    xx = a.columns.to_list()

    # TODO: s 2s 3s instead of percentiles

    for i in perc_list:
        if len(np.atleast_1d(i)) == 1:
            perc = a.apply(lambda x: np.nanpercentile(x.values, i), axis=0)
            ax.plot(xx, perc.values, color='black')
        elif len(np.atleast_1d(i)) == 2:
            perc0 = a.apply(lambda x: np.nanpercentile(x.values, i[0]), axis=0)
            perc1 = a.apply(lambda x: np.nanpercentile(x.values, i[1]), axis=0)

            ax.fill_between(xx, perc0, perc1, lw=.5, alpha=.3, color=color, **kwargs)
        else:
            raise ValueError('List items should be scalars or 2-item lists')

    ax.set_xlim(left=min(xx), right=max(xx))
    ax.set_xlabel(x)



def plot_boxplot(Load, by='day', **pltargs):
    """Return boxplot plot for each day of the week

    Arguments:
        Load (pd.Series): 1D pandas Series with timed index
        by (str): group results by 'day' or 'hour'
        **pltargs (dict): Exposes :meth:`matplotlib.pyplot.plot` arguments
    Returns:
        plot
    """
    Load = clean_convert(Load,force_timed_index=True)

    if by == 'day':
        grp = Load.groupby(Load.index.weekday)
        labels = "Mon Tue Wed Thu Fri Sat Sun".split()
    elif by == 'hour':
        grp = Load.groupby(Load.index.hour)
        labels = np.arange(0, 24)
    else:
        raise NotImplementedError('Only "day" and "hour" are implemented')
    a = []
    for __, value in grp:
        a.append(value)
    plt.boxplot(a, labels=labels, **pltargs)
    # TODO : Generalize to return monthly, hourly etc.
    # TODO Is it really needed? pd.boxplot()


def plot_LDC(Load, stacked=True, x_norm=True, y_norm=False, cmap='Spectral', color='black',
             legend=False, zoom_peak=False, ax=None, **kwargs):
    """Plot Load duration curve
    
    Arguments:
        Load (pd.Series): 1D pandas Series with timed index
        x_norm (bool): Normalize x axis (0,1)
        y_norm (bool): Normalize y axis (0,1)
        color (str): color of line. For Series only (1D)
        cmap (str): Colormap of area. For Dataframes only (2D)
        legend (bool): Show legend. For Dataframes only (2D)
        zoom_peak (bool): Show zoomed plot of peak
        kwargs (dict): exposes arguments of pd.DataFrame.plot.area
    Returns:
        Load duration curve plot
    """
    if ax is None:
        __ = plt.figure(1)
        ax_main = plt.axes()
    else:
        ax_main = ax

    if Load.ndim >= 2:
        if stacked:
            x, y = get_LDC(Load, x_norm=x_norm, y_norm=y_norm)
            # Reconverting to Dataframe as pd.plot.area is much more robust than plt.stackplot

            ldc_frame = pd.DataFrame(y, index=x, columns=Load.columns)
            y_max = np.nanmax(np.nansum(ldc_frame, axis=1))
            if y_norm: # We need to renormalize cause the sum of columns is more than 1.0
                ldc_frame = ldc_frame/ y_max
            if (ldc_frame<0).any(axis=None): # TODO: Can we stack negative values?
                raise ValueError('Cannot plot stacked area with negative values. Try stacked=False instead')
            ldc_frame.plot.area(cmap=cmap, lw=0, legend=legend, ax=ax_main, **kwargs)
        else:
            for __, v in Load.items():
                x, y = get_LDC(v, x_norm=x_norm, y_norm=y_norm)
                ax_main.plot(x, y, color=color, **kwargs)
            y_max = np.nanmax(y)
    else:
        x, y = get_LDC(Load, x_norm=x_norm, y_norm=y_norm)
        ax_main.plot(x, y, color=color)
        y_max = np.nanmax(y)

    # Set axes labels
    ax_x_min = np.min(x)
    if x_norm:
        ax_x_max = 1
        xlabel = 'Normalized duration'
    else:
        ax_x_max = len(y)
        xlabel = 'Duration'
    if y_norm:
        ax_y_max = 1
        ylabel = 'Normalized Power'
    else:
        ax_y_max = y_max
        ylabel = 'Power'

    ax_main.set_xlim(ax_x_min, ax_x_max)
    ax_main.set_xlabel(xlabel)
    ax_main.set_ylim(0, ax_y_max * 1.01)
    ax_main.set_ylabel(ylabel)
    # Draw inset plot
    if zoom_peak:
        from mpl_toolkits.axes_grid1.inset_locator import zoomed_inset_axes, mark_inset
        axins = zoomed_inset_axes(ax_main, 2.5, loc=1)
        if y.ndim >= 2:
            ldc_frame.plot.area(cmap=cmap, lw=0, legend=False, ax=axins, **kwargs)
        else:
            axins.plot(x, y, color=color)
        axins.set_xlim([ax_x_min, 0.15 * ax_x_max]) #TODO: Estimate x axis limits based on plotted values
        axins.set_ylim([0.8 * ax_y_max, ax_y_max])
        axins.get_xaxis().set_ticks([])
        axins.get_yaxis().set_ticks([])
        mark_inset(ax_main, axins, loc1=1, loc2=3, fc="none", ec="0.5")


def plot_rug(df_series, on_off=False, cmap='Greys', fig_title='', fig_width=14, normalized=False):
    """Create multiaxis rug plot from pandas Dataframe
    
    Arguments:
        df_series (pd.DataFrame): 2D pandas with timed index
        on_off (bool): if True all points that are above 0 will be plotted as one color. If False all values will be colored based on their value.
        cmap (str): colormap name (from colorbrewer, matplotlib etc.)
        fig_title (str): Figure title
        normalized (bool): if True, all series colormaps will be normalized based on the maximum value of the dataframe
    Returns:
        plot
    """
    def format_axis(iax):
        # Formatting: remove all lines (not so elegant)
        for spine in ['top','right','left','bottom']:
            iax.axes.spines[spine].set_visible(False)
            #iax.spines['right'].set_visible(False)

        # iax.xaxis.set_ticks_position('none')
        iax.yaxis.set_ticks_position('none')
        iax.get_yaxis().set_ticks([])
        iax.yaxis.set_label_coords(-.05, -.1)

    def flag_operation(v):
        if np.isnan(v) or v == 0:
            return False
        else:
            return True

    # check if Series or dataframe
    if isinstance(df_series, pd.DataFrame):
        rows = len(df_series.columns)
    elif isinstance(df_series, pd.Series):
        df_series = df_series.to_frame()
        rows = 1
    else:
        raise ValueError("Has to be either Series or Dataframe")
    if len(df_series) < 1:
        raise ValueError("Has to be non empty Series or Dataframe")

    max_color = np.nanmax(df_series.values)
    min_color = np.nanmin(df_series.values)

    __, axes = plt.subplots(nrows=rows, ncols=1, sharex=True,
                             figsize=(fig_width, 0.25 * rows), squeeze=False,
                             frameon=False, gridspec_kw={'hspace': 0.15})

    for (item, iseries), iax in zip(df_series.items(), axes.ravel()):
        format_axis(iax)
        iax.set_ylabel(str(item)[:30], rotation='horizontal',
                       rotation_mode='anchor',
                       horizontalalignment='right', x=-0.01)
        x = iseries.index

        if iseries.sum() > 0:  # if series is not empty
            if on_off:
                i_on_off = iseries.apply(flag_operation).replace(False, np.nan)
                i_on_off.plot(ax=iax, style='|', lw=.7, cmap=cmap)
            else:
                y = np.ones(len(iseries))
                #Define (truncated) colormap:
                if not normalized:  # Replace max_color (frame) with series max
                    max_color = np.nanmax(iseries.values)
                    min_color = np.nanmin(iseries.values)
                # Hack to plot max color when all series are equal
                if np.isclose(min_color, max_color):
                    min_color = min_color * 0.99

                iax.scatter(x, y,
                            marker='|', s=100,
                            c=iseries.values,
                            vmin=min_color,
                            vmax=max_color,
                            cmap=cmap)

    axes.ravel()[0].set_title(fig_title)
    axes.ravel()[-1].spines['bottom'].set_visible(True)
    axes.ravel()[-1].set_xlim(np.min(x), np.max(x))


def plot_line_holidays():
    #ax.vspan if day == holiday
    #should work only with daily
    pass

def describe_load(Load):
    """Summary plot that describes the most important features of the passed timeseries """
    pass


"""This modules contains methods which correspond to estimation of statistics (features) for timeseries."""

import numpy as np
from scipy.signal import find_peaks_cwt, ricker
from itertools import groupby

#make it work only with ndarray?

def get_mean(x, trunced=False):
    if trunced:
        x = x[x>0]
    return np.mean(x)


def get_lf(x, trunced=False):
    """Load factor"""
    if trunced:
        x = x[x>0]
    return np.mean(x)/np.max(x)

def get_trend(x, deg=1):
    # Assumes equally spaced series
    xx = np.arange(len(x))
    coeff = np.polyfit(xx, x, deg)
    return coeff[-2]

def get_rle(x, a):
    """Run length encoding algorithm.
    >>> get_rle("aaaaaaaaaaaaaaaaahhhh,,,,iuuuuuuiiaaalllaaa","a")
    [17, 3, 3]
    """
    if len(x) == 0:
        return [0]
    else:
        res = [len(list(group))
               for value, group in groupby(x)
               if value == a]
        return res if len(res) > 0 else [0]

def largest_dur_of_zero(x):
    return max(get_rle(x, 0))

def get_peaks(x, n):
    return find_peaks_cwt(x, widths=np.arange(1, n + 1), wavelet=ricker)


def get_dur_val(x, a):
    """Get duration of periods that the timeseries is above one value.
    E.g It can be used to check the non-zero load duration"""
    return len(x[x==a]) # or sum(x[x>0]) ?

def get_percentile(x, a, trunced=False):
    if trunced:
        x = x[x>0]
    return np.percentile(x, a)


def get_ramp_rates(x, a=95):
    """Retrieve ramp rate of a percentile.
    a=100 for maximum-minimum. """
    ramps = np.diff(x)
    return np.percentile(ramps, 100 - a), np.percentile(ramps, a)


def get_highest_periodicity(x): #highest peaks of fft
    from scipy.signal import welch
    length = len(x)
    w = welch(x, scaling='spectrum', nperseg=length // 2)
    peak_ind = get_peaks(w[1], 5)
    periods = [np.round(1 / w[0][peak]) for peak in peak_ind]
    #filter periods. Remove too small and too high
    max_p = length // 2
    min_p = 3  # Do not show short term AR(1) - AR(3)
    return tuple(period for period in periods if min_p < period < max_p)

def get_load_ratio(x):
    if np.min(x) != 0:
        return np.max(x) / np.min(x)

def get_autocorr(x, lag=1):
    # We redefine AR with numpy to avoid importing statsmodels
    if np.isclose(np.min(x),np.max(x)):
        #if x vector is flat, autocorrelation is not defined
        return np.nan
    else:
        return np.corrcoef(np.array([x[0:len(x)-lag], x[lag:len(x)]]))[0,1]

# TODO: 2d: parcorr

from functools import partial
## Templates of stat collection can be found below:
all_stats_desc = {'Sum': np.sum,
                  'Average': get_mean,
                  'Max': np.max,
                  'Load Factor (peakiness)': get_lf,
                  'Total Zero load duration': partial(get_dur_val, a=0),
                  'Biggest duration of consecutive zero load': lambda x: np.max(get_rle(x, a=0)),
                  'Ramps (98%)': partial(get_ramp_rates, a=98),
                  'Min (2%)': partial(get_percentile, a=0.02, trunced=False),
                  'Periodicity': lambda x: get_highest_periodicity(x)[0:2],
                  'Autocorrelation(1)': partial(get_autocorr, lag=1),
                  'Trend': get_trend,
                  'Load ratio (max/min)': get_load_ratio
                  }
#to add more...



import datetime as dt
import logging

import numpy as np
import pandas as pd

__all__ = ['make_timeseries', 'clean_convert']


def make_timeseries(x=None, year=None, length=None, startdate=None, freq=None):
    """Convert numpy array to a pandas series with a timed index. Convenience wrapper around a datetime-indexed pd.DataFrame.

    Parameters:
        x: (nd.array) raw data to wrap into a pd.Series
        startdate: pd.datetime
        year: year of timeseries
        freq: offset keyword (e.g. 15min, H)
        length: length of timeseries
    Returns:
        pd.Series or pd.Dataframe with datetimeindex
    """

    if startdate is None:
        if year is None:
            logging.info('No info on the year was provided. Using current year')
            year = dt.datetime.now().year
        startdate = dt.datetime(year, 1, 1, 0, 0, 0)

    if x is None:
        if length is None:
            raise ValueError('The length or the timeseries has to be provided')
    else:  # if x is given
        length = len(x)
        if freq is None:
            # Shortcuts: Commonly used frequencies are automatically assigned
            if len(x) == 8760 or len(x) == 8784:
                freq = 'h'
            elif len(x) == 35040:
                freq = '15min'
            elif len(x) == 12:
                freq = 'M'
            else:
                raise ValueError('Input vector length must be 12, 8760 or 35040. Otherwise freq has to be defined')

    #enddate = startdate + pd.datetools.timedelta(seconds=_freq_to_sec(freq) * (length - 1) )
    date_list = pd.date_range(start=startdate, periods=length, freq=freq)
    if x is None:
        return pd.Series(np.nan, index=date_list)
    elif isinstance(x, (pd.DataFrame, pd.Series)):
        x.index = date_list
        return x
    elif isinstance(x, (np.ndarray, list)):
        if len(x.shape) > 1:
            return pd.DataFrame(x, index=date_list)
        else:
            return pd.Series(x, index=date_list)
    else:
        raise ValueError('Unknown type of data passed')


def _freq_to_sec(freq_keyword):
    """ Converts pandas frequency string keyword to seconds. Not all frequency offsets can be converted to seconds.

    Arguments:
        freq_keyword: frequency based on pandas offsets
    Returns:
        int: corresponding seconds """
    from pandas.tseries.frequencies import to_offset

    try:
        return to_offset(freq_keyword).nanos * 1E-9
    except ValueError as e:
        raise ValueError('Works only with fixed frequencies e.g. h,s,t', e)


def human_readable_time(delta, terms=1):
    """Convert hours to human readable string

    Arguments:
        delta: time in seconds
        terms: how many word terms to use, to describe the timestep
    Returns
        str: Human readable string
    """
    # Inspired by http://stackoverflow.com/questions/26164671/convert-seconds-to-readable-format-time
    from dateutil.relativedelta import relativedelta

    intervals = ['years', 'months', 'days', 'hours', 'minutes', 'seconds']
    if delta > 31:
        delta = delta
    rd = relativedelta(hours=delta)
    out = ""
    for k in intervals[:terms]:
        if getattr(rd, k):
            out += '{} {} '.format(getattr(rd, k), k)
    return out.strip()


def clean_convert(x, force_timed_index=True, always_df=False, **kwargs):
    """Converts a list, a numpy array, or a dataframe to pandas series or dataframe, depending on the
    compatibility and the requirements. Designed for maximum compatibility.
    
    Arguments:
        x (list, np.ndarray): Vector or matrix of numbers. it can be pd.DataFrame, pd.Series, np.ndarray or list
        force_timed_index (bool): if True it will return a timeseries index
        year (int): Year that will be used for the index
        always_df (bool): always return a dataframe even if the data is one dimensional
        **kwargs: Exposes arguments of :meth:`make_timeseries`
    Returns:
        pd.Series: Timeseries
        
    """

    if isinstance(x, list):  # nice recursions
        return clean_convert(pd.Series(x), force_timed_index, always_df, **kwargs)

    elif isinstance(x, np.ndarray):
        if len(x.shape) == 1:
            return clean_convert(pd.Series(x),force_timed_index, always_df, **kwargs)
        else:
            return clean_convert(pd.DataFrame(x), force_timed_index, always_df, **kwargs)

    elif isinstance(x, pd.Series):
        if always_df:
            x = pd.DataFrame(x)
        if isinstance(x.index, pd.DatetimeIndex):
            return x
        else:  # if not datetime index
            if force_timed_index:
                logging.debug('Forcing Datetimeindex into passed timeseries.'
                      'For more accurate results please pass a pandas time-indexed timeseries.')
                return make_timeseries(x, **kwargs)
            else:  # does not require datetimeindex
                return x

    elif isinstance(x, pd.DataFrame):
        if x.shape[1] == 1 and not always_df:
            return clean_convert(x.squeeze(), force_timed_index, always_df, **kwargs)
        else:
            if force_timed_index and not isinstance(x.index, pd.DatetimeIndex):
                return make_timeseries(x, **kwargs)
            else:  # does not require datetimeindex
                return x
    else:
        raise ValueError('Unrecognized Type. Has to be one of the following: pd.DataFrame, pd.Series, np.ndarray or list')




import numpy as np
import pandas as pd
from enlopy.utils import make_timeseries
from enlopy.generate import (add_noise, gen_daily_stoch_el, gen_load_from_daily_monthly, gen_load_sinus, gen_demand_response,
                             disag_upsample, clean_convert, countweekend_days_per_month,
                             gen_analytical_LDC, gen_load_from_LDC, gen_corr_arrays, gen_gauss_markov)

class Test_noise():

    def test_ndarray_gauss(self):
        a = np.random.rand(24)
        b = np.random.rand(24) / 10
        c = gen_gauss_markov(a, b, .9)
        assert isinstance(c, np.ndarray)

    def test_ndarray_add_noise_gauss(self):
        a = np.random.rand(8760)
        b = add_noise(a, 3, 0.05)  # Gauss Markov noise
        assert isinstance(b, pd.Series)

    def test_2d_ndarray_add_noise_gauss(self):
        a = np.random.rand(8760, 2)
        b = add_noise(a, 3, 0.05)  # Gauss Markov noise
        assert isinstance(b, pd.DataFrame)
        assert (8760,2) == b.shape

    def test_ndarray_add_noise_normal(self):
        a = np.random.rand(8760)
        b = add_noise(a, 1, 0.05)  # Gauss Markov noise
        assert isinstance(b, pd.Series)

    def test_2d_ndarray_add_noise_normal(self):
        a = np.random.rand(8760, 2)
        b = add_noise(a, 1, 0.05)  # Gauss Markov noise
        assert isinstance(b, pd.DataFrame)
        assert (8760,2) == b.shape

    def test_ndarray_add_noise_uniform(self):
        a = np.random.rand(8760)
        b = add_noise(a, 2, 0.05)  # Gauss Markov noise
        assert isinstance(b, pd.Series)

    def test_2d_ndarray_add_noise_uniform(self):
        a = np.random.rand(8760, 2)
        b = add_noise(a, 2, 0.05)  # Gauss Markov noise
        assert isinstance(b, pd.DataFrame)
        assert (8760,2) == b.shape

    def test_add_noise_not_annual(self):
        a = np.random.rand(15)
        b = add_noise(a, 3, 0.05)
        assert isinstance(b, pd.Series)


class Test_gen_monthly_daily():

    def test_gen_monthly_daily(self):
        Weight = .55  # Ratio between working and non-working day load (e.g. 70% - 30% )
        ML = 1000 * np.ones(12)  # monthly load
        DWL = np.random.rand(24) * 10  # daily load working
        DWL = DWL / DWL.sum()  # normalized
        DNWL = np.random.rand(24) * 5  # daily load non working
        DNWL = DNWL / DNWL.sum()  # daily load non working
        year = 2014
        Load1 = gen_load_from_daily_monthly(ML, DWL, DNWL, Weight, year)
        assert len(Load1) == 8760
        assert np.isclose(Load1.sum(), np.sum(ML))

class Test_gen_dummy_load():
    def test_gen_dummy(self):
        a = gen_daily_stoch_el(1500)
        assert isinstance(a, np.ndarray)
        assert len(a) == 24

class Test_gen_load_sinus():
    def test_gen_sinus(self):
        Load1 = gen_load_sinus(1,2,3,4,5,6)
        assert len(Load1) == 8760

class Test_disag():
    def test_disag_daily_to_hourly(self):
        x = np.arange(0, 365)
        y = (np.cos(2 * np.pi / 364 * x) * 50 + 100)

        y = make_timeseries(y, freq='D')

        disag_profile = np.random.rand(24)
        y_disag = disag_upsample(y, disag_profile)
        assert np.isclose(np.sum(y_disag), np.sum(y)) # <= 0.001 #FIXME: np test equality
        assert len(y_disag) == 8760

    def test_disag_hourly_to_minutes(self):
        x = np.arange(0, 8760)
        y = (np.cos(2 * np.pi / 8759 * x) * 50 + 100)
        y = make_timeseries(y, freq='h')

        disag_profile = np.random.rand(60)
        y_disag = disag_upsample(y, disag_profile, to_offset='min')
        assert np.isclose(np.sum(y_disag), np.sum(y) ) # <= 0.001 #FIXME: np test equality
        assert len(y_disag) == 8760*60


class Test_demand_side_management():

    def test_load_shifting_small(self):
        a = np.random.rand(8760) * 100
        a = clean_convert(a, force_timed_index=True, always_df=False)
        b = gen_demand_response(a,.1,.2)
        assert np.isclose(np.sum(a), np.sum(b))
        assert np.max(a) > np.max(b)
    def test_load_shifting_big(self):
        a = np.random.rand(8760) * 100
        a = clean_convert(a, force_timed_index=True, always_df=False)
        b = gen_demand_response(a,.15,.5)
        assert np.isclose(np.sum(a), np.sum(b))
        assert np.max(a) > np.max(b)

def test_countweekend_days_per_month():
    a = make_timeseries(year=2015, length=8760, freq='h')
    b = countweekend_days_per_month(a.resample('d').mean())
    assert len(b) == 12
    assert sum(b) == 104 #weekend days in 2015

def test_gen_analytical_LDC():
    #Generate a simple LDC with Peak of 1
    U = (1, 0.5, 0.2, 8760)
    LDC = gen_analytical_LDC(U)

    assert max(LDC[0]) == 1.0
    assert min(LDC[0]) == 0.0
    assert np.isclose(np.mean(LDC[0]), 0.5)


def test_gen_load_from_LDC():
    # Only operate 90% of the time.
    duration_fraction = 0.9
    LDC = gen_analytical_LDC((1, 0.5, 0.2, 8760 * duration_fraction))
    b = gen_load_from_LDC(LDC)
    assert b.max() <= 1.0

    # According to the defined formula anything below should be zero
    val_perc = np.percentile(b, (1 - duration_fraction - 0.01) * 100)
    assert np.isclose(val_perc, 0.0)


def test_gen_corr_arrays():
    Na = 2
    length = 1000
    r = 0.85
    M = np.array([[1, r],
                  [r, 1]])
    A = gen_corr_arrays(Na, length, M)
    new_r = np.corrcoef(A)[0][1]
    assert A.shape == (Na, length)
    #allow some tolerance of convergence..
    assert np.abs(new_r - r) <= 0.03


import numpy as np
import pandas as pd
from enlopy.utils import make_timeseries
from enlopy.analysis import reshape_timeseries, get_LDC, get_load_stats


def test_reshape_timeseries():
    a = np.random.rand(8760)
    b = reshape_timeseries(a, x='dayofyear', y='hour')
    assert b.shape == (24,365)
    assert np.isclose(b.sum().sum(), a.sum())

def test_reshape_timeseries_week():
    a = np.random.rand(8760)
    b = reshape_timeseries(a, x='hour', y='week', aggfunc='mean')
    assert b.shape == (52,24)
    assert np.isclose(b.mean().mean(), a.mean(),0.001)

def test_reshape_multiannual():
    a = np.random.rand(8760*2)
    a = make_timeseries(a, year=2019, freq='h')
    b = reshape_timeseries(a, x='dayofyear', y='hour', aggfunc='sum')
    assert b.shape == (24,365)
    assert np.isclose(b.sum().sum(), a.sum())

def test_reshape_multiannual_dataframe():
    a = np.random.rand(8760*2)
    a = pd.DataFrame(make_timeseries(a, year=2019, freq='h'))
    b = reshape_timeseries(a, x='dayofyear', y='hour', aggfunc='sum')
    assert b.shape == (24,365)
    assert np.isclose(b.sum().sum(), a.sum())

def test_reshape_monthly_multiannual():
    ind = pd.DatetimeIndex(pd.date_range(start='Jan-1990', end='Jan-2017', freq='M'))
    a = pd.DataFrame(np.random.rand(324), index=ind)
    b = reshape_timeseries(a, x='month', y='year', aggfunc='sum')
    assert b.shape == (27, 12)  # 27 years, 12 months
    assert np.isclose(b.sum().sum(), a.sum())

def test_get_LDC():
    a = np.random.rand(8760*2)
    a = make_timeseries(a, freq='h')
    b = get_LDC(a)
    assert np.isclose(b[1].sum(), a.sum())
    #check monotonicity
    assert np.all(np.diff(b[1]) < 0)

def test_get_LDC_not_annual():
    a = np.random.rand(10000)
    a = make_timeseries(a, freq='h')
    b = get_LDC(a)
    assert np.isclose(b[1].sum(), a.sum())
    #check monotonicity
    assert np.all(np.diff(b[1]) < 0)

def test_get_LDC_not_annual_ndarray():
    a = np.random.rand(10000)
    b = get_LDC(a)
    assert np.isclose(b[1].sum(), a.sum())
    #check monotonicity
    assert np.all(np.diff(b[1]) < 0)

def test_get_LDC_2d():
    a = np.random.rand(8760, 4)
    a = make_timeseries(a, freq='h')
    b = get_LDC(a)
    assert np.isclose(np.nansum(b[1]), np.nansum(a))
    # check monotonicity
    assert np.all(np.diff(b[1].sum(1)) < 0)

def test_get_stats():
    a = np.ones(8760)
    a = make_timeseries(a, freq='h')
    b = get_load_stats(a)
    assert a.sum() == b.loc['Sum'].squeeze()
    assert np.asarray(b.loc['Ramps (98%)'])[0] == (0,0)
    assert np.isclose(0, b.loc['Trend'].squeeze())
    assert 1 == b.loc['Load Factor (peakiness)'].squeeze()

def test_get_stats_df():
    a = np.ones(8760)
    a = make_timeseries(a, freq='h').to_frame()
    b = get_load_stats(a)
    assert a.sum().sum() == b.loc['Sum'].squeeze()



""" Setup script for enlopy"""

import codecs
from setuptools import setup, find_packages
import os
import re

# import distutils.command.bdist_conda


HERE = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

requirements = ['numpy>=1.10',
                'scipy>=1.10',
                'matplotlib>=2.2',
                'pandas>=2.0']

setup(
    name="enlopy",
    author="Konstantinos Kavvadias",
    author_email="kavvkon@gmail.com",
    url='https://github.com/kavvkon/enlopy',
    description='Python library with methods to generate, process, analyze, and plot energy related timeseries.',
    long_description=read('README.rst'),
    license="BSD-3-Clause",
    version=find_version("enlopy", "__init__.py"),
    python_requires = ">=3.9",
    install_requires=requirements,
    keywords=['energy','timeseries','statistics','profile','demand'],
    packages=find_packages(),
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        "Topic :: Utilities",
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development :: Libraries :: Python Modules',

        'Natural Language :: English',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',

        #    distclass=distutils.command.bdist_conda.CondaDistribution,
        #    conda_buildnum=1,
        #    conda_features=['mkl'],

    ],

)
